from tkinter import simpledialog;
import importlib.resources as pkg_resources;
import time;
import threading, sys, os, ctypes;

from pycman_dna.map import Map
from pycman_dna.map  import Direction;

class PacmanGame:
    """
    PacmanGame's base constructor method. Performs initial setup.
    """
    __update_rate = 0.25;
    __moves_taken = 0;
    def __init__(self):
        self.__screen = None
        self.__game_running = True;
        self.__message = "No message"
        self.__map = None;
        self.__map_name = ""
        self.set_update_rate(250)
        ctypes.windll.shcore.SetProcessDpiAwareness(2)


    def __run_game(self) -> None:
        """
        
        """
        import pygame;
        pygame.init();
        
        self.__screen = pygame.display.set_mode([784, 784], pygame.SCALED)
        #self.__screen = pygame.display.set_mode([600, 600], pygame.SCALED)
        pygame.display.set_icon(pygame.image.load(os.path.dirname(__file__) + "\\images\\icon.png"))
        pygame.display.set_caption("Pycman - " + "\"" +self.__map_name + "\"")

        while True:
            for event in pygame.event.get():
                if event.type == pygame.QUIT or (event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE):
                    self.__game_running = False
                    os._exit(1)
            # Fill the background with black
            self.__screen.fill((0, 0, 0))
            
            self.__map.draw_map(self.__screen);
            
            # Flip the display
            pygame.display.flip()
        pygame.quit() 
        

    def __check_end_conditions(self):
        if self.__map._Map__dot_count <= 0:
            self.__game_running = False
            self.send_message("All dots collected - well done!")
        elif self.__moves_taken >= self.__map._Map__total_moves:
            self.__game_running = False
            self.send_message("No moves remaining. Please try again.")
            self.__map.update_pacman([0,0], Direction.DEAD)
        elif self.__map.get_pacman_direction() == Direction.DEAD:
            self.__game_running = False
            self.send_message("Pacman has been caught. Please try again.")


    def set_map(self, map_name : str) -> None:
        """
        Sets the game map that Pacman will travel around, based on the map name passed in.
        :param map_name: The name of a built-in game map.
        """
        self.__map_name = map_name
        self.__map = Map(map_name.lower(), self)
        pass;


    def show_game(self) ->None:
        """
        Brings up the game window containing the chosen game map.
        *Please ensure set_map() has been called first.*
        """
        if self.__map is not None:
            self.__thread = threading.Thread(target=PacmanGame.__run_game, args=[self]);
            self.__thread.start();

            time.sleep(1)
        else:
            print("MapError: Map not set. Please set with set_map() method first.")


    def move(self) -> None:
        """
        Moves Pacman one space in the direction he is facing. Counts as a move.
        """
        if(self.__game_running):
            delta_movement = [0, 0]
            match self.__map.get_pacman_direction():
                case Direction.UP:
                    delta_movement[1] = -1
                case Direction.DOWN:
                    delta_movement[1] = 1
                case Direction.LEFT:
                    delta_movement[0] = -1
                case Direction.RIGHT:
                    delta_movement[0] = 1

            self.__map.update_pacman(delta_movement, self.__map.get_pacman_direction())
            PacmanGame.__moves_taken += 1;

            self.__check_end_conditions()
            time.sleep(PacmanGame.__update_rate);


    def turn_right(self) -> None:
        """
        Turns Pacman 90 degrees to the right. Does not count as a move.
        """
        if self.__game_running:
            new_direction = Direction.UP
            match self.__map.get_pacman_direction():
                case Direction.UP:
                    new_direction = Direction.RIGHT
                case Direction.DOWN:
                    new_direction = Direction.LEFT
                case Direction.LEFT:
                    new_direction = Direction.UP
                case Direction.RIGHT:
                    new_direction = Direction.DOWN
            self.__map.update_pacman([0,0], new_direction)
            time.sleep(PacmanGame.__update_rate);

    def turn_left(self) -> None:
        """
        Turns Pacman 90 degrees to the left. Does not count as a move.
        """
        if self.__game_running:
            new_direction = Direction.UP
            match self.__map.get_pacman_direction():
                case Direction.UP:
                    new_direction = Direction.LEFT
                case Direction.DOWN:
                    new_direction = Direction.RIGHT
                case Direction.LEFT:
                    new_direction = Direction.DOWN
                case Direction.RIGHT:
                    new_direction = Direction.UP
            self.__map.update_pacman([0,0], new_direction)
            time.sleep(PacmanGame.__update_rate);

    
    def send_message(self, message : str) -> None:
        """
        Sends a decimal number message to the game window.
        :param message: The text message that should be printed.
        """
        self.__message = message
    

    def is_obstacle_ahead(self, obstacle : str) -> bool:
        """
        Checks whether or not the requested obstacle is directly in front of Pacman. Takes the current direction of Pacman into account. Possible options are:
        \n - Empty
        \n - Dot
        \n - Wall
        \n - Pill
        \n - Void
        \n - Ghost\n
        :param obstacle: The name of an obstacle to look for. Parameter is case insensitive.
        :return: A boolean value (true/false) based on whether or not the requested obstacle was found.
        """
        return obstacle.lower() == self.get_obstacle_ahead().lower()


    def get_obstacle_ahead(self) -> str:
        """
        Returns the name of the obstacle directly in front of Pacman. Takes the current direction of Pacman into account.
        :return: A string containing the name of the obstacle directly ahead of Pacman.
        """
        player_pos = self.__map._Map__player.map_index
        match self.__map._Map__player.direction:
            case Direction.UP:
                return self.__map.get_obstacle_at_coords(player_pos[0], player_pos[1] - 1).name
            case Direction.DOWN:
                return self.__map.get_obstacle_at_coords(player_pos[0], player_pos[1] + 1).name
            case Direction.LEFT:
                return self.__map.get_obstacle_at_coords(player_pos[0] - 1, player_pos[1]).name
            case Direction.RIGHT:
                return self.__map.get_obstacle_at_coords(player_pos[0] + 1, player_pos[1]).name
            
        return self.__map.get_obstacle_at_coords(player_pos[0], player_pos[1]).name


    def set_update_rate(self, update_rate : int) -> None:
        """
        Sets the delay between each update of the game. A lower number results in a faster update rate. 
        \nValues are automatically clamped between 75 and 1000 milliseconds.
        :param update_rate: The amount of time, in milliseconds, to wait between updates.
        """
        if update_rate < 75:
            update_rate = 75
        elif update_rate > 1000:
            update_rate = 1000

        PacmanGame.__update_rate = update_rate * 0.001


    def get_text_from_user(self, prompt: str) -> str:
        """
        Opens an input field for the user to enter some text by keyboard. Pauses the game until complete.
        :param prompt: A message prompt to let the user know what kind of text to enter.
        :return: A string containing the user's entered text.
        """
        text = simpledialog.askstring("Pacman", prompt);
        return text if text is not None else "";


    def get_int_from_user(self, prompt: str) -> int:
        """
        Opens an input field for the user to enter some whole number by keyboard. Pauses the game until complete.
        :param prompt: A message prompt to let the user know what kind of number to enter.
        :return: An integer containing the user's entered number. Returns -1 if something goes wrong.
        """
        num = simpledialog.askinteger("Pacman", prompt);
        return num if num is not None else -1;


    def get_float_from_user(self, prompt: str) -> float:
        """
        Opens an input field for the user to enter some decimal number by keyboard. Pauses the game until complete.
        :param prompt: A message prompt to let the user know what kind of number to enter.
        :return: A double containing the user's entered number. Returns -1 if something goes wrong.
        """
        num = simpledialog.askfloat("Pacman", prompt);
        return num if num is not None else -1;

    def print_map_names(self) -> None:
        """
        Prints out all the names of the built-in game maps available to load.\n
        Uses the standard output stream, which is the terminal for Python projects.
        """
        print("Valid Map Names:\n----------------")
        names = pkg_resources.read_text('pycman_dna', 'map_names.txt').split()
        counter = 0
        for name in names:
            print(str(counter) + ": " + name)
            counter+=1

        
        


    


import pygame;
import os
from enum import Enum;
from typing import List;

class Obstacle(Enum):
    EMPTY = 1
    DOT = 2
    WALL = 3
    GHOST = 4
    PILL = 5
    PACMAN = 6
    VOID = 7


class Direction(Enum):
    UP = 1
    DOWN = 2
    LEFT = 3
    RIGHT = 4
    DEAD = 5


class Map:
    __MAP_HEIGHT = 20;
    __MAP_WIDTH = 20;


    def __init__(self, map_name : str, game) -> None:
        self.__game = game;
        self.__board_model = [['' for x in range(Map.__MAP_HEIGHT)] for y in range(Map.__MAP_WIDTH)]
        self.__board_view = [[pygame.Surface((32,32), pygame.SRCALPHA, 32) for x in range(Map.__MAP_HEIGHT)] for y in range(Map.__MAP_WIDTH)]
        pygame.font.init()
        self.__font = pygame.font.Font(pygame.font.get_default_font(), 24)
        self.__dot_count = 0
        self.__player = Pacman()
        self.__ghost = None;
        self.__build_map_model(map_name)


    def __build_map_model(self, map_name : str) -> None:
        map_data = self.__read_map(map_name)
        self.__difficulty = map_data[0][3:];
        self.__total_moves = int(map_data[1]);
        for i in range(0, len(map_data) - 2):
            chars = list(map_data[i + 2])
            for j in range(0, len(chars)):
                match chars[j]:
                    case '-':
                        self.__board_model[j][i] = Obstacle.EMPTY
                    case '.':
                        self.__dot_count += 1
                        self.__board_model[j][i] = Obstacle.DOT
                    case 'g':
                        self.__board_model[j][i] = Obstacle.GHOST
                        self.__ghost = Ghost()
                        self.__ghost.set_map_index(j, i)
                        self.__ghost.sprite = load_image("ghost.png");
                    case 'x':
                        self.__board_model[j][i] = Obstacle.PILL
                    case 'p':
                        self.__board_model[j][i] = Obstacle.PACMAN
                        self.__player.set_map_index(j, i)
                        self.__player.sprite = load_image("PacmanUP.png");
                    case 'w':
                        self.__board_model[j][i]  = Obstacle.WALL

                self.__build_map_view(i, j)  
    

    def __build_map_view(self, column : int, row : int) -> None:
        obstacleName = self.__board_model[row][column].name;
        if obstacleName.upper() != "PACMAN" and obstacleName.upper() != "GHOST" and obstacleName.upper() != "EMPTY":
            self.__board_view[row][column] = load_image(obstacleName.lower() + ".png")
        

    def draw_map(self, screen : pygame.Surface) -> None:
        pygame.draw.rect(screen, (255,255,255), rect=pygame.Rect(0, 0, 784, 50))
        screen.blit(self.__font.render(self.__game._PacmanGame__message, True, (0,0,0)), (8, 8))

        for i in range(0, self.__MAP_WIDTH):
            for j in range(0, self.__MAP_HEIGHT):
                if self.__board_model[j][i] != Obstacle.EMPTY and self.__board_model[j][i] != Obstacle.PACMAN and self.__board_model[j][i] != Obstacle.GHOST:
                    screen.blit(self.__board_view[j][i], [54 + (j * 32), 54 + (i * 32)])

        if(self.__player is not None):
            screen.blit(self.__player.sprite, self.__player.position)
        if(self.__ghost is not None):
            screen.blit(self.__ghost.sprite, self.__ghost.position)

        pygame.draw.rect(screen, (255,255,255), rect=pygame.Rect(0, 734, 784, 50))
        screen.blit(self.__font.render("difficulty:" + str(self.__difficulty) + ", player position [" + str(self.__player.map_index[0]) + ", " + str(self.__player.map_index[1]) +"], steps:" + str(self.__game._PacmanGame__moves_taken) + "/" + str(self.__total_moves), True, (0,0,0)), (8, 748))


    def update_pacman(self, delta_movement, new_direction : Direction) -> None:
        if delta_movement != [0,0]:
            new_position = [0, 0]
            new_position[0] = self.__player.map_index[0] + delta_movement[0];
            new_position[1] = self.__player.map_index[1] + delta_movement[1];
            if new_position[0] >= self.__MAP_WIDTH or new_position[1] >= self.__MAP_HEIGHT or new_position[0] < 0 or new_position[1] < 0:
                raise Exception("Index was outside the bounds of the array.")
            if self.get_obstacle_at_coords(new_position[0], new_position[1]) == Obstacle.DOT:
                self.__dot_count -= 1
            elif self.get_obstacle_at_coords(new_position[0], new_position[1]) == Obstacle.PILL:
                self.__ghost._Ghost__toggle_edible_state()
            
            if self.get_obstacle_at_coords(new_position[0], new_position[1]) != Obstacle.WALL:
                self.__board_model[self.__player.map_index[0]][self.__player.map_index[1]] = Obstacle.EMPTY
                self.__player.set_map_index(new_position[0], new_position[1])
                self.__board_model[self.__player.map_index[0]][self.__player.map_index[1]] = Obstacle.PACMAN

                self.update_ghost()

            self.__board_view[new_position[0]][new_position[1]] = None

        if self.__player.direction != Direction.DEAD:
            self.__player.set_direction(new_direction)

    
    def update_ghost(self):
        if self.__ghost is not None:
            if self.__ghost.is_edible and self.__player.map_index == self.__ghost.map_index:
                self.__ghost = None
            
            if self.__ghost is not None:
                delta_movement = [0, 0]
                if self.__player.map_index[1] < self.__ghost.map_index[1]:
                    delta_movement[1] = -1
                elif self.__player.map_index[0] < self.__ghost.map_index[0]:
                    delta_movement[0] = -1
                elif self.__player.map_index[1] > self.__ghost.map_index[1]:
                    delta_movement[1] = 1
                elif self.__player.map_index[0] > self.__ghost.map_index[0]:
                    delta_movement[0] = 1

                self.__ghost.set_map_index(self.__ghost.map_index[0] + delta_movement[0], self.__ghost.map_index[1] + delta_movement[1])

                if self.__ghost.is_edible == False and self.__player.map_index == self.__ghost.map_index:
                    self.__player.set_direction(Direction.DEAD)
                    self.__ghost.set_map_index(self.__ghost.map_index[0] - delta_movement[0], self.__ghost.map_index[1] - delta_movement[1])
                



    def get_obstacle_at_coords(self, column : int, row : int) -> Obstacle:
        if row < 0 or column < 0 or row >= self.__MAP_HEIGHT or column >= self.__MAP_WIDTH:
            return Obstacle.VOID
        else:
            return self.__board_model[column][row]

    def __read_map(self, map_name : str) -> List[str]:
        lines = []
        with open(get_map_path(map_name + '.map')) as f:
            lines = f.read().splitlines()
        return lines;
    
    def get_pacman(self) -> 'Pacman':
        return self.__player;

    def get_pacman_direction(self) -> Direction:
        return self.__player.direction;


def load_image(package_path: str) -> pygame.Surface:
        return pygame.image.load(os.path.dirname(__file__) + "\\images\\" + package_path)

def get_map_path(package_path: str) -> List[str]:
    return os.path.dirname(__file__) + "\\maps\\" + package_path

class Pacman:
    def __init__(self) -> None:
        self.map_index = [0, 0]
        self.position = []
        self.sprite_list = []
        self.create_sprite_list()
        self.set_direction(Direction.UP)
        self.sprite = pygame.Surface((32,32));

    def set_map_index(self, row : int, column : int) -> None:
        self.map_index = [row, column]
        self.position = [54 + (row * 32), 54 + (column * 32)]

    def create_sprite_list(self):
        for i in range(5):
            self.sprite_list.append(load_image("Pacman" + Direction(i + 1).name + ".png"))
            #self.sprite_list.append(pygame.image.load("images/Pacman" + Direction(i + 1).name + ".png"))
        self.sprite = self.sprite_list[0]

    def set_direction(self, new_direction : Direction):
        self.direction = new_direction
        self.sprite = self.sprite_list[new_direction.value - 1]

class Ghost:
    def __init__(self) -> None:
        self.is_edible = False;
        self.map_index = [0, 0];
        self.position = [];
        self.sprite = pygame.Surface((32,32));

    def set_map_index(self, row : int, column : int) -> None:
        self.map_index = [row, column]
        self.position = [54 + (row * 32), 54 + (column * 32)]

    def toggle_edible_state(self) -> None:
        self.is_edible = not self.is_edible

        if self.is_edible:
            self.sprite = load_image("eatable_ghost.png")
        else:
            self.sprite = load_image("ghost.png")

    
